#include <gtest/gtest.h>
#include <platform.h>

#include "util.h"


#define TEST_COMPILE(src) DO(test_compile(src, "expr"))

#define TEST_ERRORS_1(src, err1) \
  { const char* errs[] = {err1, NULL}; \
    DO(test_expected_errors(src, "expr", errs)); }

#define TEST_ERRORS_2(src, err1, err2) \
  { const char* errs[] = {err1, err2, NULL}; \
    DO(test_expected_errors(src, "expr", errs)); }


class ArrayTest : public PassTest
{};


TEST_F(ArrayTest, AsVal)
{
  const char* src =
    "trait val T\n"
    "primitive P1 is T\n"
    "primitive P2 is T\n"
    "primitive P3 is T\n"

    "primitive Foo\n"
    "  fun apply(): Array[T] =>\n"
    "    [as T: P1; P2; P3]";

  TEST_COMPILE(src);
}


TEST_F(ArrayTest, AsIso)
{
  const char* src =
    "trait iso T\n"
    "class iso C1 is T\n"
    "class iso C2 is T\n"
    "class iso C3 is T\n"

    "primitive Foo\n"
    "  fun apply(): Array[T] =>\n"
    "    [as T: C1; C2; C3]";

  TEST_COMPILE(src);
}


TEST_F(ArrayTest, AsIsoAliasedElements)
{
  const char* src =
    "trait iso T\n"
    "class iso C1 is T\n"
    "class iso C2 is T\n"
    "class iso C3 is T\n"

    "primitive Foo\n"
    "  fun apply(): Array[T] =>\n"
    "    let c1: C1 = C1\n"
    "    let c2: C2 = C2\n"
    "    let c3: C3 = C3\n"
    "    [as T: c1; c2; c3]";

  TEST_ERRORS_1(src, "array element not a subtype of specified array type");
}


TEST_F(ArrayTest, AsIsoConsumedElements)
{
  const char* src =
    "trait iso T\n"
    "class iso C1 is T\n"
    "class iso C2 is T\n"
    "class iso C3 is T\n"

    "primitive Foo\n"
    "  fun apply(): Array[T] =>\n"
    "    let c1: C1 iso = C1\n"
    "    let c2: C2 iso = C2\n"
    "    let c3: C3 iso = C3\n"
    "    [as T: consume c1; consume c2; consume c3]";

  TEST_COMPILE(src);
}


TEST_F(ArrayTest, NoAsVal)
{
  const char* src =
    "trait val T\n"
    "primitive P1 is T\n"
    "primitive P2 is T\n"
    "primitive P3 is T\n"

    "primitive Foo\n"
    "  fun apply(): Array[(P1 | P2 | P3)] =>\n"
    "    [P1; P2; P3]";

  TEST_COMPILE(src);
}


TEST_F(ArrayTest, NoAsIso)
{
  const char* src =
    "trait iso T\n"
    "class iso C1 is T\n"
    "class iso C2 is T\n"
    "class iso C3 is T\n"

    "primitive Foo\n"
    "  fun apply(): Array[(C1 | C2 | C3)] =>\n"
    "    [C1; C2; C3]";

  TEST_COMPILE(src);
}


TEST_F(ArrayTest, NoAsIsoAliasedElements)
{
  const char* src =
    "trait iso T\n"
    "class iso C1 is T\n"
    "class iso C2 is T\n"
    "class iso C3 is T\n"

    "primitive Foo\n"
    "  fun apply(): Array[(C1 | C2 | C3)] =>\n"
    "    let c1: C1 = C1\n"
    "    let c2: C2 = C2\n"
    "    let c3: C3 = C3\n"
    "    [c1; c2; c3]";

  TEST_ERRORS_1(src,
    "array element not a subtype of specified array type");
}


TEST_F(ArrayTest, NoAsIsoConsumedElements)
{
  const char* src =
    "trait iso T\n"
    "class iso C1 is T\n"
    "class iso C2 is T\n"
    "class iso C3 is T\n"

    "primitive Foo\n"
    "  fun apply(): Array[(C1 | C2 | C3)] =>\n"
    "    let c1: C1 iso = C1\n"
    "    let c2: C2 iso = C2\n"
    "    let c3: C3 iso = C3\n"
    "    [consume c1; consume c2; consume c3]";

  TEST_COMPILE(src);
}


TEST_F(ArrayTest, InferFromArrayType)
{
  const char* src =
    "trait iso T\n"
    "class iso C1 is T\n"
    "class iso C2 is T\n"
    "class iso C3 is T\n"

    "primitive Foo\n"
    "  fun apply(): Array[T] =>\n"
    "    [C1; C2; C3]";

  TEST_COMPILE(src);
}


TEST_F(ArrayTest, InferFromInterfaceWithTypeParam)
{
  const char* src =
    "trait iso T\n"
    "class iso C1 is T\n"
    "class iso C2 is T\n"
    "class iso C3 is T\n"

    "interface Arrayish[A]\n"
    "  fun apply(index: USize): this->A ?\n"

    "primitive Foo\n"
    "  fun apply(): Arrayish[T] =>\n"
    "    [C1; C2; C3]";

  TEST_COMPILE(src);
}


TEST_F(ArrayTest, InferFromSeq)
{
  const char* src =
    "trait iso T\n"
    "class iso C1 is T\n"
    "class iso C2 is T\n"
    "class iso C3 is T\n"

    "primitive Foo\n"
    "  fun apply(): Seq[T] =>\n"
    "    [C1; C2; C3]";

  TEST_COMPILE(src);
}


TEST_F(ArrayTest, InferFromApplyInterface)
{
  const char* src =
    "trait iso T\n"
    "class iso C1 is T\n"
    "class iso C2 is T\n"
    "class iso C3 is T\n"

    "interface ArrayishOfT\n"
    "  fun apply(index: USize): this->T ?\n"

    "primitive Foo\n"
    "  fun apply(): ArrayishOfT =>\n"
    "    [C1; C2; C3]";

  TEST_COMPILE(src);
}


TEST_F(ArrayTest, InferFromApplyInterfaceValElementNoArrowThis)
{
  const char* src =
    "trait val T\n"
    "primitive P1 is T\n"
    "primitive P2 is T\n"
    "primitive P3 is T\n"

    "interface ArrayishOfT\n"
    "  fun apply(index: USize): T ?\n"

    "primitive Foo\n"
    "  fun apply(): ArrayishOfT =>\n"
    "    [P1; P2; P3]";

  TEST_COMPILE(src);
}


TEST_F(ArrayTest, InferFromApplyInterfaceIsoElementNoArrowThis)
{
  const char* src =
    "trait iso T\n"
    "class iso C1 is T\n"
    "class iso C2 is T\n"
    "class iso C3 is T\n"

    "interface ArrayishOfT\n"
    "  fun apply(index: USize): T ?\n"

    "primitive Foo\n"
    "  fun apply(): ArrayishOfT =>\n"
    "    [C1; C2; C3]";

  TEST_ERRORS_1(src, "function body isn't the result type");
}


TEST_F(ArrayTest, InferFromApplyInterfaceBoxElementInTuple)
{
  // From issue #2233
  const char* src =
    "trait box T\n"
    "class box C1 is T\n"
    "  new box create() => None\n"
    "class box C2 is T\n"
    "  new box create() => None\n"
    "class box C3 is T\n"
    "  new box create() => None\n"

    "interface ArrayishOfTAndU64\n"
    "  fun apply(index: USize): this->(T, U64) ?\n"

    "primitive Foo\n"
    "  fun apply(): ArrayishOfTAndU64 =>\n"
    "    [(C1, 1); (C2, 2); (C3, 3)]";

  TEST_COMPILE(src);
}


TEST_F(ArrayTest, InferFromValuesInterface)
{
  const char* src =
    "trait iso T\n"
    "class iso C1 is T\n"
    "class iso C2 is T\n"
    "class iso C3 is T\n"

    "interface ArrayishOfT\n"
    "  fun values(): Iterator[this->T]^\n"

    "primitive Foo\n"
    "  fun apply(): ArrayishOfT =>\n"
    "    [C1; C2; C3]";

  TEST_COMPILE(src);
}


TEST_F(ArrayTest, InferFromValuesOfArrayInterface)
{
  const char* src =
    "trait iso T\n"
    "class iso C1 is T\n"
    "class iso C2 is T\n"
    "class iso C3 is T\n"

    "interface ArrayishOfArrayOfT\n"
    "  fun values(): Iterator[Array[T]]^\n"

    "primitive Foo\n"
    "  fun apply(): ArrayishOfArrayOfT =>\n"
    "    [[C1]; [C2]; [C3]]";

  TEST_COMPILE(src);
}


TEST_F(ArrayTest, InferFromReadSeqAndReadElementInterface)
{
  const char* src =
    "interface box ReadSeq[A]\n"
    "  fun apply(i: USize): this->A ?\n"
    "interface box ReadElement[A]\n"
    "  fun apply(i: USize): A ?\n"

    "primitive Foo\n"
    "  fun apply(): (ReadSeq[U8] & ReadElement[U8]) =>\n"
    "    [1; 2; 3]";

  TEST_COMPILE(src);
}


TEST_F(ArrayTest, InferFromValuesCallWithIteratorAntecedent)
{
  const char* src =
    "primitive Foo\n"
    "  fun apply(): (Iterator[U8] | None) =>\n"
    "    [1; 2; 3].values()";

  TEST_COMPILE(src);
}


TEST_F(ArrayTest, InferFromArrayIso)
{
  const char* src =
    "trait iso T\n"
    "class iso C1 is T\n"
    "class iso C2 is T\n"
    "class iso C3 is T\n"

    "primitive Foo\n"
    "  fun apply(): Array[T] iso^ =>\n"
    "    [C1; C2; C3]";

  TEST_COMPILE(src);
}


TEST_F(ArrayTest, InferFromArrayVal)
{
  const char* src =
    "trait val T\n"
    "primitive P1 is T\n"
    "primitive P2 is T\n"
    "primitive P3 is T\n"

    "primitive Foo\n"
    "  fun apply(): Array[T] val =>\n"
    "    [P1; P2; P3]";

  TEST_COMPILE(src);
}


TEST_F(ArrayTest, InferFromArrayIsoWithAs)
{
  const char* src =
    "trait iso T\n"
    "class iso C1 is T\n"
    "class iso C2 is T\n"
    "class iso C3 is T\n"

    "primitive Foo\n"
    "  fun apply(): Array[T] iso^ =>\n"
    "    [as T: C1; C2; C3]";

  TEST_COMPILE(src);
}


TEST_F(ArrayTest, InferFromArrayValWithAs)
{
  const char* src =
    "trait val T\n"
    "primitive P1 is T\n"
    "primitive P2 is T\n"
    "primitive P3 is T\n"

    "primitive Foo\n"
    "  fun apply(): Array[T] val =>\n"
    "    [as T: P1; P2; P3]";

  TEST_COMPILE(src);
}


TEST_F(ArrayTest, InferFromArrayIsoMustBeRecovered)
{
  const char* src =
    "trait T\n"
    "class C1 is T\n"
    "class C2 is T\n"
    "class C3 is T\n"

    "primitive Foo\n"
    "  fun apply(): Array[T] iso^ =>\n"
    "    let c1: C1 = C1\n"
    "    let c2: C2 = C2\n"
    "    let c3: C3 = C3\n"
    "    [c1; c2; c3]";

  TEST_ERRORS_1(src, "array element not a subtype of specified array type");
}


TEST_F(ArrayTest, InferFromArrayValMustBeRecovered)
{
  const char* src =
    "trait T\n"
    "class C1 is T\n"
    "class C2 is T\n"
    "class C3 is T\n"

    "primitive Foo\n"
    "  fun apply(): Array[T] val =>\n"
    "    let c1: C1 = C1\n"
    "    let c2: C2 = C2\n"
    "    let c3: C3 = C3\n"
    "    [c1; c2; c3]";

  TEST_ERRORS_1(src, "array element not a subtype of specified array type");
}


TEST_F(ArrayTest, InferFromArrayIsoMayBeExplicitlyRecovered)
{
  const char* src =
    "trait T\n"
    "class C1 is T\n"
    "class C2 is T\n"
    "class C3 is T\n"

    "primitive Foo\n"
    "  fun apply(): Array[T] iso^ =>\n"
    "    recover\n"
    "      let c1: C1 = C1\n"
    "      let c2: C2 = C2\n"
    "      let c3: C3 = C3\n"
    "      [c1; c2; c3]\n"
    "    end";

  TEST_COMPILE(src);
}


TEST_F(ArrayTest, InferFromArrayValMayBeExplicitlyRecovered)
{
  const char* src =
    "trait T\n"
    "class C1 is T\n"
    "class C2 is T\n"
    "class C3 is T\n"

    "primitive Foo\n"
    "  fun apply(): Array[T] val =>\n"
    "    recover\n"
    "      let c1: C1 = C1\n"
    "      let c2: C2 = C2\n"
    "      let c3: C3 = C3\n"
    "      [c1; c2; c3]\n"
    "    end";

  TEST_COMPILE(src);
}


TEST_F(ArrayTest, InferFromArrayBoxNeedNotRecovered)
{
  const char* src =
    "trait T\n"
    "class C1 is T\n"
    "class C2 is T\n"
    "class C3 is T\n"

    "primitive Foo\n"
    "  fun apply(): Array[T] box =>\n"
    "    let c1: C1 = C1\n"
    "    let c2: C2 = C2\n"
    "    let c3: C3 = C3\n"
    "    [c1; c2; c3]";

  TEST_COMPILE(src);
}


TEST_F(ArrayTest, EmptyArrayLiteral)
{
  const char* src =
    "trait T\n"

    "class Foo\n"
    "  let a1: Array[T] = []\n"
    "  let a2: (U64 | Array[T] | None) = []";

  TEST_COMPILE(src);
}


TEST_F(ArrayTest, EmptyArrayLiteralMustBeInferable)
{
  const char* src =
    "trait T1\n"
    "trait T2\n"

    "primitive Foo\n"
    "  fun apply() =>\n"
    "    let a1 = []\n"
    "    let a2: (Array[T1] | Array[T2]) = []";

  TEST_ERRORS_2(src,
    "must specify the element type or it must be inferable",
    "must specify the element type or it must be inferable");
}
